﻿/*
====== READ THIS ========
This code was created as a part of the seminar on 
flash game programing at the University of Munich (Ludwig-
Maximlians Universität.

The code may be used for educational purposes only. All
rights reserverd by the author.

Author: Max Maurer
E-Mail: max.maurer@ifi.lmu.de
(c) 2011
If you have any questions don't hasitate to contact me.
=========================
*/
package {
	import flash.display.MovieClip;
	import flash.events.Event;
	import flash.text.TextField;
	public class Field extends MovieClip {
		private var currentTile:Tile;
		private var level:int;
		private var score:int;
		private var frame:int;
		private var moveDownEvery:int;
		private var fieldBlocks:MyMatrix;
		private const MAX_HEIGHT:int = 16;
		private const MAX_WIDTH:int = 7;
		private var linesRemoved:int;
		private var blocks:MovieClip;
		private var particles:MovieClip;
		private var emitAtOnce:int;
		public function Field():void {
			addEventListener(Event.ENTER_FRAME, enterFrameHandler);
			resetGame();
			emitAtOnce=4;
		}
		
		public function resetGame():void {
			if (blocks!=null) while (blocks.numChildren>0) blocks.removeChildAt(0);
			if (particles!=null) while (particles.numChildren>0) particles.removeChildAt(0);
			score=0;
			level=1;
			frame=0;
			moveDownEvery=60;
			fieldBlocks=new MyMatrix();
			linesRemoved=0;
			createTile();
			updateUi();
			blocks = new MovieClip();
			addChild(blocks);
			particles = new MovieClip();
			addChild(particles);
			Key.setKeyRepeat(38,15); // only repeat rotation every 15 frames
			Key.setKeyRepeat(37,15); // only repeat left every 15 frames
			Key.setKeyRepeat(39,15); // only repeat right every 15 frames
			Key.setKeyRepeat(40,15); // only repeat down every 5 frames
		}
		
		private function createTile():void {
			if (currentTile!=null) {
				removeChild(currentTile);
			}
			currentTile = new Tile();
			addChild(currentTile);
		}
		
		private function enterFrameHandler(evt:Event):void {
			var lost:Boolean=false;
			frame++;
			Key.nextFrame();
			if (Key.isDown(38)) currentTile.rotate();
			else if (Key.isDown(37)) {
				if (currentTile.tx>0 && !isBlocked(-1)) currentTile.moveLeft();
			} else if (Key.isDown(39)) {
				if (currentTile.tx<MAX_WIDTH-currentTile.getWidth() &&!isBlocked(+1)) currentTile.moveRight();
			} else if (Key.isDown(40)) {
				lost= checkCollission();
				checkLines();
				currentTile.moveDown();
			}
			if (frame%moveDownEvery==0) {
				lost=checkCollission();
				checkLines();
				// do this last to remove tiles with the next tick
				currentTile.move();
			}
			// animate particles
			for (var i:int=0;i<particles.numChildren;i++) {
				var p:Particle = Particle(particles.getChildAt(i));
				if (p.isAlive()) p.animate();
				else particles.removeChild(p);
			}
			if (lost) {
				resetGame();
			}
		}
		
		private function isBlocked(dx:int):Boolean {
			// check for all pieces vertically of x if dx is empty
			trace(currentTile.getHeight());
			trace(currentTile.getWidth());
			for (var ty:int=0;ty<currentTile.getHeight();ty++) {
				for (var tx:int=0;tx<currentTile.getWidth();tx++) {
					var b:Block = currentTile.getBlock(tx,ty);
					if (b==null) continue;
					if (fieldBlocks.getObject(currentTile.tx+tx+dx,currentTile.ty+ty)!=null) {
						return true;
					}
				}
			}
			return false;
		}
		
		private function checkCollission():Boolean {
			// check if tile is on bottom and integrati
			// for each tile of the tile check if the next field ist blocked or if the next field is bottom
			var stopTile:Boolean = false;
			var ty:int;
			var tx:int;
			var b:Block;
			for (ty=0;ty<currentTile.getHeight();ty++) {
				for (tx=0;tx<currentTile.getWidth();tx++) {
					b = currentTile.getBlock(tx,ty);
					if (b==null) continue;
					if (fieldBlocks.getObject(currentTile.tx+tx,currentTile.ty+ty+1)!=null) {
						// we have a tile below us
						stopTile=true;
					}
					if (currentTile.ty+ty>=MAX_HEIGHT-1) {
						stopTile=true;
					}
				}
			}
			// integrate the tile if necessary
			var lost:Boolean = false; // when the tile is added on top of another tile the player has lost 
			if (stopTile) {
				for (ty=0;ty<currentTile.getHeight();ty++) {
					for (tx=0;tx<currentTile.getWidth();tx++) {
						b = currentTile.getBlock(tx,ty);
						if (b==null) continue;
						if (addBlock(currentTile.tx+tx,currentTile.ty+ty,b)) {
							lost=true;
						}
					}
				}
				createTile();
			}
			return lost;
		}
		
		private function addBlock(x:int,y:int,b:Block):Boolean {
			blocks.addChild(b);
			b.x = x*b.width;
			b.y = y*b.height;
			return fieldBlocks.setObject(x,y,b);
		}
		
		private function checkLines():void {
			// remove full lines
			var linesToRemove:Array = new Array();
			var y:int;
			for (y=MAX_HEIGHT-1;y>=0;y--) {
				if (checkLine(y)) linesToRemove.push(y);
			}
			// add score
			switch (linesToRemove.length) {
				case 1:
				score+=100;
				break;
				case 2:
				score+=200;
				break;
				case 3:
				score+=400;
				break;
				case 4:
				score+=1000;
				break;
			}
			linesRemoved+=linesToRemove.length;
			updateUi();
			// rebuild the field omitting those lines
			var currentLine=MAX_HEIGHT-1;
			var newField = new MyMatrix();
			for (y=MAX_HEIGHT-1;y>=0;y--) {
				if (linesToRemove.length>0 && y==linesToRemove[0]) {
					removeLine(linesToRemove.shift());
					continue;
				}
				for (var x:int=0;x<MAX_WIDTH;x++) {
					var b:Block = Block(fieldBlocks.getObject(x,y));
					if (b==null) continue;
					b.y = currentLine*b.height;
					newField.setObject(x,currentLine,b);
				}
				currentLine--;
			}
			fieldBlocks=newField;
		}
		
		private function removeLine(y:int):void {
			for (var x:int=0;x<MAX_WIDTH;x++) {
				var b:Block = Block(fieldBlocks.getObject(x,y));
				if (b==null) continue; // should never happen
				// emit debris at this place
				for (var j:int=0;j<emitAtOnce;j++) {
					particles.addChild(new Particle(b.x+b.width/2,b.y+b.height/2,b.getColor()));
				}
				blocks.removeChild(b);
			}
		}
		
		private function checkLine(y:int):Boolean {
			for (var x:int=0;x<MAX_WIDTH;x++) {
				if (fieldBlocks.getObject(x,y)==null) return false;
			}
			return true;
		}

		private function updateUi():void {
			txtScore.text = ""+score;
			level = Math.floor(linesRemoved/10)+1;
			moveDownEvery = 10+50/level;
			txtLevel.text = ""+level;
		}

	}
}